#include "sshthread.h"
#include <errno.h>

static int my_channel_read(ssh_channel channel, void *dest, uint32_t count,
		int is_stderr)
{
	ssh_channel ch[2] = { channel, NULL };
	ssh_channel outch[2] = { channel, NULL };
	fd_set readfds;
	struct timeval tv;
	int ret;

	FD_ZERO(&readfds);
	for (;;) {
		tv.tv_sec = SSHTHREAD_TIMEOUT;
		tv.tv_usec = 0;
		ret = ssh_select(ch, outch, 0, &readfds, &tv);
		if (ret == SSH_EINTR)
			continue;
		if (ret == SSH_ERROR) {
			if (outch[0] == NULL)
				return 0;
		}
		ret = channel_read_nonblocking(channel, dest, count,
				is_stderr);
		if (!ret && !ssh_channel_is_eof(channel))
			continue;
		return ret;
	}
}

static int my_channel_read_in_full(ssh_channel channel, void *dest,
		unsigned int count, int is_stderr)
{
	char *d = (char *) dest;
	int total = 0;
	int ret;

	while (count > 0) {
		ret = my_channel_read(channel, d, count, is_stderr);
		if (ret < 0) {
			if (total)
				return total;
			return ret;
		}
		if (!ret)
			return total;
		d += ret;
		total += ret;
		count -= ret;
	}

	return total;
}

SshThread::SshThread(QString public_key_file, QString private_key_file,
		QString host, int port, unsigned long serial)
{
	public_key_path = public_key_file;
	private_key_path = private_key_file;
	remote_host = host;
	remote_host_port = port;
	validation_status = NOT_STARTED;
	serial_number = serial;
}

SshThread::~SshThread() {
}

void SshThread::run()
{
	int verbosity = SSH_LOG_NOLOG;
	int timeout = SSHTHREAD_TIMEOUT;
	int rc;
	char buffer[256];
	char *end;
	unsigned int total;
	int nbytes;
	int public_key_type = 0;
	int private_key_type = 0;
	unsigned long serial;
	ssh_private_key private_key;
	ssh_string public_key;

	session = ssh_new();
	validation_status = INIT_FAILED;
	if (!session)
		goto out;

	ssh_options_set(session, SSH_OPTIONS_HOST,
			remote_host.toLocal8Bit().constData());
	ssh_options_set(session, SSH_OPTIONS_USER, USER_NAME);
	ssh_options_set(session, SSH_OPTIONS_LOG_VERBOSITY, &verbosity);
	ssh_options_set(session, SSH_OPTIONS_PORT, &remote_host_port);
	ssh_options_set(session, SSH_OPTIONS_TIMEOUT, &timeout);

	emit verStatus(QString::fromUtf8("Łączenie z rejestratorem"));

	rc = ssh_connect(session);
	validation_status = CONNECTION_FAILED;
	if (rc != SSH_OK)
		goto out_free;

	emit verStatus(QString::fromUtf8("Autoryzacja"));

	public_key = publickey_from_file(session,
			public_key_path.toLocal8Bit().constData(),
			&public_key_type);
	validation_status = PUBLIC_KEY_FAILED;
	if (!public_key)
		goto out_disconnect;

	rc = ssh_userauth_offer_pubkey(session, NULL, public_key_type,
			public_key);
	validation_status = AUTH_FAILED;
	if (rc == SSH_AUTH_ERROR || rc == SSH_AUTH_DENIED)
		goto out_disconnect;

	private_key = privatekey_from_file(session,
			private_key_path.toLocal8Bit().constData(),
			private_key_type, NULL);
	validation_status = PRIVATE_KEY_FAILED;
	if (!private_key)
		goto out_disconnect;

	rc = ssh_userauth_pubkey(session, NULL, public_key, private_key);
	if (rc == SSH_AUTH_ERROR || rc == SSH_AUTH_DENIED) {
		validation_status = AUTH_FAILED;
		goto out_disconnect;
	} else if (rc == SSH_AUTH_PARTIAL) {
		validation_status = AUTH_PARTIAL;
		goto out_disconnect;
	}

	channel = channel_new(session);
	validation_status = CHANNEL_INIT_FAIL;
	if (!channel)
		goto out_disconnect;

	rc = channel_open_session(channel);
	validation_status = SESSION_INIT_FAIL;
	if (rc != SSH_OK)
		goto out_channel_free;

	emit verStatus(QString::fromUtf8("Weryfikacja numeru rejestratora"));

	rc = channel_request_exec(channel, "get-serial-no");
	validation_status = COMMAND_ERROR;
	if (rc != SSH_OK)
		goto out_channel_close;
	channel_send_eof(channel);

	nbytes = my_channel_read_in_full(channel, buffer,
			sizeof(buffer) - 1, 0);
	channel_close(channel);
	channel_free(channel);

	validation_status = CHENNEL_END;
	if (nbytes < 0)
		goto out_disconnect;
	buffer[nbytes] = '\0';

	serial = strtol(buffer, &end, 0);
	validation_status = PARSE_ERROR;
	if ((end == buffer) || (*end != '\0' && *end != '\n'))
		goto out_disconnect;

	validation_status = SERIAL_MISMATCH;
	if (serial != serial_number)
		goto out_disconnect;

	channel = channel_new(session);
	validation_status = CHANNEL_INIT_FAIL;
	if (channel == NULL)
		goto out_disconnect;

	rc = channel_open_session(channel);
	validation_status = SESSION_INIT_FAIL;
	if (rc != SSH_OK)
		goto out_channel_free;

	emit verStatus(QString
			::fromUtf8("Weryfikacja oprogramowania rejestratora"));

	rc = channel_request_exec(channel, "verify-firmware");
	validation_status = COMMAND_ERROR;
	if (rc != SSH_OK)
		goto out_channel_close;
	channel_send_eof(channel);

	/* FIXME */
	/* this is based on the assumption that neither PASSED nor
	 * FAILED string will occur only if the session is
	 * terminated (timeouts) */

	validation_status = SESSION_TIMEOUT;
	total = 0;
	for (;;) {
		char *next, *s;
		char verify_firmware_str[] = "verify-firmware: ";

		/* read want to process one line per iteration */
		for (next = buffer; next != buffer + total; next++) {
			if (*next == '\r' || *next == '\n')
				break;
		}
		if (next == buffer + total) {
			if (total == sizeof(buffer) - 1)
				total = 0;
			nbytes = my_channel_read(channel, buffer + total,
					sizeof(buffer) - 1 - total, 0);
			if (nbytes <= 0)
				break;

			total += nbytes;
			buffer[total] = '\0';
			continue;
		}
		*next = '\0';
		next++;

		s = strstr(buffer, "Result: PASSED");
		if (s) {
			validation_status = SUCCESS;
			break;
		}
		s = strstr(buffer, "Result: FAILED");
		if (s) {
			validation_status = FIRMWARE_CORRUPTED;
			break;
		}

		s = strstr(buffer, verify_firmware_str);
		if (s) {
			int progress;

			s += sizeof(verify_firmware_str) - 1;
			progress = strtol(s, &end, 10);
			if (*end == '%')
				emit verProgress(progress);
		}
		memmove(buffer, next, total - (next - buffer));
		total = total - (next - buffer);
	}

out_channel_close:
	channel_close(channel);
out_channel_free:
	channel_free(channel);
out_disconnect:
	ssh_disconnect(session);
out_free:
	ssh_free(session);
out:
	return;
}

void SshThread::closeChannel()
{
	channel_send_eof(channel);
	channel_close(channel);
	channel_free(channel);
}

enum validation_status SshThread::getValidationStatus()
{
	return validation_status;
}
