/*
 * uff-utils - uff.h
 *
 * UFF file format description
 *
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <uff.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

typedef uint64_t uff64;
typedef uint32_t uff32;
typedef uint16_t uff16;
typedef uint8_t uff8;

#define UFF_MAGIC		0x11b67c2c34a20000ULL

#ifdef HAVE_BYTESWAP_H
#include <byteswap.h>
#else
#define bswap_16(x) \
	((((x) >> 8) & 0xff) | (((x) & 0xff) << 8))

#define bswap_32(x) \
	((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >> 8) |	\
	(((x) & 0x0000ff00) << 8) | (((x) & 0x000000ff) << 24))

#define bswap_64(x) \
	((((x) & 0xff00000000000000ull) >> 56)				\
	| (((x) & 0x00ff000000000000ull) >> 40)				\
	| (((x) & 0x0000ff0000000000ull) >> 24)				\
	| (((x) & 0x000000ff00000000ull) >> 8)				\
	| (((x) & 0x00000000ff000000ull) << 8)				\
	| (((x) & 0x0000000000ff0000ull) << 24)				\
	| (((x) & 0x000000000000ff00ull) << 40)				\
	| (((x) & 0x00000000000000ffull) << 56))
#endif
#include <sys/param.h>

#define uff8_to_cpu(x)	(x)
#define uffi8_to_cpu(x)	(x)
#define cpu_to_uff8(x)	(x)
#define cpu_to_uffi8(x)	(x)

#if BYTE_ORDER == BIG_ENDIAN
#define uff16_to_cpu(x)	(x)
#define uffi16_to_cpu(x)	(x)
#define uff32_to_cpu(x)	(x)
#define uffi32_to_cpu(x)	(x)
#define uff64_to_cpu(x)	(x)
#define uffi64_to_cpu(x)	(x)
#define cpu_to_uff16(x)	(x)
#define cpu_to_uffi16(x)	(x)
#define cpu_to_uff32(x)	(x)
#define cpu_to_uffi32(x)	(x)
#define cpu_to_uff64(x)	(x)
#define cpu_to_uffi64(x)	(x)
#elif BYTE_ORDER == LITTLE_ENDIAN
#define uff16_to_cpu(x)	(bswap_16(x))
#define uffi16_to_cpu(x)	((int16_t) bswap_16(x))
#define uff32_to_cpu(x)	(bswap_32(x))
#define uffi32_to_cpu(x)	((int32_t) bswap_32(x))
#define uff64_to_cpu(x)	(bswap_64(x))
#define uffi64_to_cpu(x)	((int64_t) bswap_64(x))
#define cpu_to_uff16(x)	(bswap_16(x))
#define cpu_to_uffi16(x)	((int16_t) bswap_16(x))
#define cpu_to_uff32(x)	(bswap_32(x))
#define cpu_to_uffi32(x)	((int32_t) bswap_32(x))
#define cpu_to_uff64(x)	(bswap_64(x))
#define cpu_to_uffi64(x)	((int64_t) bswap_64(x))
#else
#error "Please define ar3cmonitor protocol conversion macros"
#endif

#ifndef EOVERFLOW
#define EOVERFLOW	512
#endif

struct uff_section_desc {
	uff64 offset;
	uff64 size;
} __attribute__((packed));

struct uff_header {
	uff64 magic;
	uff16 version;
	uff16 revision;
	uff32 serial;
} __attribute__((packed));

struct uff_section {
	unsigned long present:1;
	unsigned long size;
	void *data;
};

struct uff {
	int mapped;
	unsigned long serial;
	struct uff_section sects[UFF_SECTION_MAX];
};


uff_t uff_alloc(void)
{
	struct uff *uff;
	unsigned int i;

	uff = malloc(sizeof(*uff));
	if (uff != NULL) {
		uff->serial = 0;
		for (i = 0; i < UFF_SECTION_MAX; i++) {
			uff->sects[i].size = 0;
			uff->sects[i].data = NULL;
		}
	}
	return uff;
}

static int uff_destroy_section(struct uff *uff, unsigned int section)
{
	free(uff->sects[section].data);
	uff->sects[section].data = NULL;
	return 0;
}

void uff_destroy(uff_t uff)
{
	unsigned int i;

	for (i = 0; i < UFF_SECTION_MAX; i++)
		uff_destroy_section(uff, i);
}

void uff_free(uff_t uff)
{
	uff_destroy(uff);
	free(uff);
}

static void *uff_map(struct uff *uff, FILE *f, unsigned long offset,
		unsigned long size)
{
	int ret;
	size_t readed;
	void *p;

	if (offset > LONG_MAX)
		return NULL;
	if (size > SIZE_MAX)
		return NULL;

	ret = fseek(f, offset, SEEK_SET);
	if (ret)
		return NULL;

	p = malloc(size);
	if (p == NULL)
		return NULL;

	readed = fread(p, size, 1, f);
	if (!readed) {
		free(p);
		return NULL;
	}
	return p;
}

static int uff_load_section(struct uff *uff, const struct uff_section_desc *d,
		FILE *f, unsigned int section)
{
	struct uff_section *s;
	uint64_t offset;
	uint64_t size;

	s = &uff->sects[section];
	offset = uff64_to_cpu(d->offset);
	size = uff64_to_cpu(d->size);
	if (size > ULONG_MAX)
		return -EINVAL;
	if (offset > ULONG_MAX)
		return -EINVAL;

	if (!size)
		return 0;

	s->size = size;
	s->data = uff_map(uff, f, offset, size);
	if (s->data == NULL)
		return -EINVAL;
	return 0;
}

int uff_load_f(uff_t uff, FILE *f)
{
	struct uff_header hdr;
	struct uff_section_desc d[UFF_SECTION_MAX];
	size_t readed;
	int ret;
	unsigned int i;

	readed = fread(&hdr, sizeof(hdr), 1, f);
	if (!readed)
		return -EINVAL;

	if (uff16_to_cpu(hdr.version) > 0)
		return -EINVAL;

	uff->serial = uff32_to_cpu(hdr.serial);

	readed = fread(&d, sizeof(d[0]), UFF_SECTION_MAX, f);
	if (!readed)
		return -EINVAL;

	for (i = 0; i < readed; i++) {
		ret = uff_load_section(uff, &d[i], f, i);
		if (ret)
			goto out_destroy_sections;
	}
	return 0;

out_destroy_sections:
	for (i = 0; i < UFF_SECTION_MAX; i++)
		uff_destroy_section(uff, i);
	return ret;
}

#define uff_pad(x)	(((x) + 4095) & ~4095)

int uff_store_f(uff_t uff, FILE *f)
{
	struct uff_header hdr;
	struct uff_section_desc d[UFF_SECTION_MAX];
	unsigned long pos;
	unsigned long size;
	unsigned int i;
	size_t written;

	hdr.magic = cpu_to_uff64(UFF_MAGIC);
	hdr.version = cpu_to_uff16(0);
	hdr.revision = cpu_to_uff16(0);
	hdr.serial = cpu_to_uff32(uff->serial);
	written = fwrite(&hdr, sizeof(hdr), 1, f);
	if (!written)
		return -errno;

	pos = uff_pad(sizeof(hdr) + sizeof(d));
	for (i = 0; i < UFF_SECTION_MAX; i++) {
		unsigned long old_pos;

		size = uff->sects[i].size;
		if (!size) {
			d[i].offset = cpu_to_uff64(0);
			d[i].size = cpu_to_uff64(0);
			continue;
		}

		d[i].offset = cpu_to_uff64(pos);
		d[i].size = cpu_to_uff64(size);

		old_pos = pos;
		pos = uff_pad(pos + size);

		/*
		 * size + padding is always <= ULONG_MAX + 1
		 * and in worst case the new position will
		 * overflow to old position
		 */
		if (pos <= old_pos)
			return -EOVERFLOW;

		/* fseek() limit */
		if (pos >= LONG_MAX)
			return -EOVERFLOW;
	}
	written = fwrite(d, sizeof(d), 1, f);
	if (!written)
		return -errno;

	for (i = 0; i < UFF_SECTION_MAX; i++) {
		size = uff->sects[i].size;
		if (!size)
			continue;

		pos = uff64_to_cpu(d[i].offset);
		fseek(f, pos, SEEK_SET);

		written = fwrite(uff->sects[i].data, size, 1, f);
		if (!written)
			return -errno;
	}
	return 0;
}

int uff_load(struct uff *uff, const char *file)
{
	int ret;
	FILE *f;

	f = fopen(file, "rb");
	if (f == NULL)
		return -errno;
	ret = uff_load_f(uff, f);
	fclose(f);
	return ret;
}

int uff_store(struct uff *uff, const char *file)
{
	int ret;
	FILE *f;

	f = fopen(file, "wb");
	if (f == NULL)
		return -errno;
	ret = uff_store_f(uff, f);
	fclose(f);
	return ret;
}

unsigned long uff_get_serial(struct uff *uff)
{
	return uff->serial;
}

int uff_set_serial(struct uff *uff, unsigned long serial)
{
	uff->serial = serial;
	return 0;
}

size_t uff_section_size(struct uff *uff, unsigned int section)
{
	if (section >= UFF_SECTION_MAX)
		return 0;
	return uff->sects[section].size;
}

void *uff_section_map(struct uff *uff, unsigned int section)
{
	if (section >= UFF_SECTION_MAX)
		return NULL;
	return uff->sects[section].data;
}

ssize_t uff_section_read(uff_t uff, unsigned int section, void *buf,
		size_t count)
{
	if (section >= UFF_SECTION_MAX)
		return 0;

	if (count > uff->sects[section].size)
		count = uff->sects[section].size;

	memcpy(buf, uff->sects[section].data, count);
	return count;
}

ssize_t uff_section_write(uff_t uff, unsigned int section, void *buf,
		size_t count)
{
	void *p;

	if (section >= UFF_SECTION_MAX)
		return 0;
	if (count >= ULONG_MAX)
		return -EOVERFLOW;

	p = malloc(count);
	if (p == NULL)
		return -EINVAL;

	memcpy(p, buf, count);
	free(uff->sects[section].data);
	uff->sects[section].data = p;
	uff->sects[section].size = count;
	return count;
}

int uff_section_from_file(uff_t uff, unsigned int section, const char *fname)
{
	FILE *f;
	void *p = NULL;
	size_t size = 4096;
	size_t len = 0;
	size_t readed;
	int ret;

	if (section >= UFF_SECTION_MAX)
		return 0;

	f = fopen(fname, "rb");
	if (f == NULL)
		return -errno;

	while (!feof(f)) {
		size_t old_len;

		ret = -ENOMEM;
		if (len >= size)
			size *= 2;

		p = realloc(p, size);
		if (p == NULL)
			goto out_close;

		readed = fread((char *) p + len, 1, size - len, f);
		old_len = len;
		len += readed;
		ret = -EOVERFLOW;
		if (len < old_len)
			goto out_close;
	}
	ret = -EOVERFLOW;
	if (len >= ULONG_MAX)
		goto out_close;

	ret = 0;
	p = realloc(p, len);

	free(uff->sects[section].data);
	uff->sects[section].data = p;
	uff->sects[section].size = len;

out_close:
	fclose(f);

	if (ret)
		free(p);
	return ret;
}

int uff_section_to_file(uff_t uff, unsigned int section, const char *fname)
{
	return -ENOSYS;
}
